#!/usr/bin/env python
# coding: utf-8
"""
Configuration and helpers for the tests of pangres with pytest.
"""
import pandas as pd
import json
from sqlalchemy import create_engine

# # Helpers

def drop_table_if_exists(engine, schema, table_name):
    namespace = f'{schema}.{table_name}' if schema is not None else table_name
    engine.execute(f'DROP TABLE IF EXISTS {namespace};')


# ## Class TestDB

# +
def pytest_addoption(parser):
    parser.addoption('--sqlite_conn', action="store")
    parser.addoption('--pg_conn', action="store")
    parser.addoption('--mysql_conn', action="store")
    parser.addoption('--pg_schema', action='store', default=None)

def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".
    conn_strings = {'sqlite':metafunc.config.option.sqlite_conn,
                    'pg':metafunc.config.option.pg_conn,
                    'mysql':metafunc.config.option.mysql_conn}
    engines = []
    schemas = []
    for db_type, conn_string in conn_strings.items():
        if conn_string is None:
            raise ValueError(f'Missing connection string for database type: {db_type}')
        schema = metafunc.config.option.pg_schema if db_type == 'pg' else None
        engine = create_engine(conn_string)
        schemas.append(schema)
        engines.append(engine)
    assert len(engines) == len(schemas)
    metafunc.parametrize("engine, schema", list(zip(engines, schemas)), scope='module')


# -

# ## Function to read back from database data we inserted
# We need to apply a few modification for comparing DataFrames we get back from the DB and DataFrames we expect e.g. for JSON (with SQlite pandas reads it as string).

def read_example_table_from_db(engine, schema, table_name):
    def load_json_if_needed(obj):
        """
        For SQlite we receive strings back (or None) for a JSON column.
        For Postgres we receive lists or dicts (or None) back.
        """
        if isinstance(obj, str):
            return json.loads(obj)
        return obj
    namespace = f'{schema}.{table_name}' if schema is not None else table_name
    df_db = (pd.read_sql(f'SELECT * FROM {namespace}', con=engine, index_col='profileid')
             .astype({'likes_pizza':bool})
             .assign(timestamp=lambda df: pd.to_datetime(df['timestamp'], utc=True))
             .assign(favorite_colors= lambda df: df['favorite_colors'].map(load_json_if_needed)))
    return df_db



#!/usr/bin/env python
# coding: utf-8
import json
from math import floor
from sqlalchemy import VARCHAR
from pangres import upsert
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import drop_table_if_exists


# # Test data


# +
df = _TestsExampleTable.create_example_df(nb_rows=20000)

# pandas can't handle JSON
# so for testing the speed of pd.to_sql we need to cast list and dicts to str
json_like_to_str = lambda x: json.dumps(x) if isinstance(x, (dict, list)) else x
df_no_json_like = df.assign(favorite_colors=lambda df: df['favorite_colors'].map(json_like_to_str))


# -


# # Helpers

# +
# table for speed test with pangres
pangres_table_name = 'test_speed_pangres'
# table for speed test with pd.to_sql
pandas_table_name = 'test_speed_pandas_create_table'

def create_or_update(engine, schema, if_row_exists):
    # MySQL does not want flexible text length in indices/PK
    dtype={'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    upsert(engine=engine, df=df, schema=schema,
           table_name=pangres_table_name, if_row_exists=if_row_exists,
           dtype=dtype)

def pd_to_sql(engine, schema):
    # MySQL does not want flexible text length in indices/PK
    dtype={'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    # we need this for SQlite when using pandas table creation
    # since we cannot use more than 999 parameters in a parameterized query
    if 'sqlite' in engine.dialect.dialect_description:
        col_len = len(df.columns) + len(df.index.names)
        chunksize = floor(999 / col_len)
    else:
        chunksize=None
    
    df_no_json_like.to_sql(con=engine, schema=schema,
                           name=pandas_table_name,
                           method='multi', chunksize=chunksize,
                           dtype=dtype)
# -

# # Creation speed


# do all tests only once because:
#
# 1. we have enough data as is
# 2. we would have to drop and recreate table at each loop...

def test_creation_speed_with_upsert(engine, schema, benchmark):
    # benchmark is implicitly imported with pytest
    drop_table_if_exists(engine=engine, schema=schema, table_name=pangres_table_name)
    benchmark.pedantic(lambda: create_or_update(engine=engine, schema=schema, if_row_exists='update'),
                       rounds=1, iterations=1)


# # Upsert overwrite speed

def test_upsert_overwrite_speed_with_upsert(engine, schema, benchmark):
    benchmark.pedantic(lambda: create_or_update(engine=engine, schema=schema, if_row_exists='update'),
                       rounds=1, iterations=1)


# # Upsert keep speed

def test_upsert_keep_speed_with_upsert(engine, schema, benchmark):
    benchmark.pedantic(lambda: create_or_update(engine=engine, schema=schema, if_row_exists='ignore'),
                       rounds=1, iterations=1)


# # Compare with pandas

def test_creation_speed_with_pandas(engine, schema, benchmark):
    drop_table_if_exists(engine=engine, schema=schema, table_name=pandas_table_name)
    benchmark.pedantic(lambda: pd_to_sql(engine=engine, schema=schema),
                       rounds=1, iterations=1)



#!/usr/bin/env python
# coding: utf-8
"""
This module tests if uploading data in chunks works as
expected (we should get the correct DataFrame length back).
"""
import pandas as pd
from sqlalchemy import VARCHAR
from pangres import upsert
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import read_example_table_from_db, drop_table_if_exists


# # Helpers

def insert_chunks(engine, schema, chunksize, nb_rows):
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)
    table_name=f'test_insert_chunksize_{chunksize}'
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    upsert(schema=schema,
           table_name=table_name,
           df=df,
           chunksize=chunksize,
           engine=engine,
           if_row_exists='update',
           # MySQL does not want flexible text length in indices/PK
           dtype={'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None)
    df_db = read_example_table_from_db(engine=engine, schema=schema, table_name=table_name)
    # sort index (for MySQL...)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


# # Tests

# ## Insert values one by one

def test_insert_one(engine, schema):
    insert_chunks(engine, schema, chunksize=1, nb_rows=11)


# ## Insert an odd size of chunks that is not a multiple of df length

def test_insert_odd_chunksize(engine, schema):
    insert_chunks(engine, schema, chunksize=3, nb_rows=11)




# +
import logging
import os

loggers = {}

def log(text, name='pangres', level=logging.INFO):
    """
    Logs given text to stderr (default of the logging library).

    Parameters
    ----------
    text : str
        Text to log
    name : str, default "pangres"
        Name of the logger
    level : int, default logging.INFO

    Notes
    -----
    This is heavily inspired from:
    https://github.com/SergeyPirogov/webdriver_manager/blob/master/webdriver_manager/logger.py

    Examples
    --------
    >>> log('info')

    >>> import logging
    >>> log('warning!', level=logging.WARNING)
    """
    # environment variable so user can customize the logging level of pangres
    log_level = os.getenv('PANGRES_LOG_LEVEL')
    if log_level:
        level = int(log_level)
    # only add a handler if we do not have one (otherwise we'd have duplicated logs)
    if loggers.get(name):
        loggers.get(name).info(text)
    else:
        _logger = logging.getLogger(name)
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s | %(levelname)s     '
                                      '| %(name)s    | %(module)s:%(funcName)s:%(lineno)s '
                                      '- %(message)s')
        handler.setFormatter(formatter)
        _logger.addHandler(handler)
        _logger.setLevel(level)
        loggers[name] = _logger
        _logger.info(text)



#!/usr/bin/env python
# coding: utf-8
"""
Here we want to test if:
1. the IGNORE/UPDATE logic works properly
2. NULL values do not cause issues
3. Crappy text insert and column names does not cause issues
"""
import pandas as pd
import random
from sqlalchemy import VARCHAR
from pangres import upsert, fix_psycopg2_bad_cols
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import read_example_table_from_db, drop_table_if_exists


# # Config

table_name = 'test_upsert'
default_args = {'table_name':table_name,
                'create_schema':True,
                'add_new_columns':True,
                'adapt_dtype_of_empty_db_columns':False}


# # Test data

# +
df = _TestsExampleTable.create_example_df(nb_rows=5)
# test for NULL values except for boolean column
df.iloc[0,[ix for ix, col in enumerate(df.columns) if col != 'likes_pizza']] = None

# test for update
df2 = _TestsExampleTable.create_example_df(nb_rows=6)

# test for ignore
df3 = _TestsExampleTable.create_example_df(nb_rows=6)


# -

# # Tests
# ORDER MATTERS!

# ## Create table

def test_create_table(engine, schema):
    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    upsert(engine=engine, schema=schema, df=df, if_row_exists='update', dtype=dtype, **default_args)
    df_db = read_example_table_from_db(engine=engine, schema=schema, table_name=table_name)
    pd.testing.assert_frame_equal(df, df_db)


# ## INSERT UPDATE 

def test_upsert_update(engine, schema):
    dtype = {'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None

    upsert(engine=engine, schema=schema, df=df2, if_row_exists='update', dtype=dtype, **default_args)
    df_db = read_example_table_from_db(engine=engine, schema=schema, table_name=table_name)
    pd.testing.assert_frame_equal(df2, df_db)


# ## INSERT IGNORE

def test_upsert_ignore(engine, schema):
    dtype = {'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    
    drop_table_if_exists(engine=engine, schema=schema, table_name=table_name)
    for _df in (df, df3):
        upsert(engine=engine, schema=schema, df=_df, if_row_exists='ignore', dtype=dtype, **default_args)
    df_db = read_example_table_from_db(engine=engine, schema=schema, table_name=table_name)
    expected = pd.concat((df, df3.tail(1)), axis=0)
    pd.testing.assert_frame_equal(expected, df_db)


# # Add colums with crappy names and insert crappy text values

def test_crappy_text_insert(engine, schema):
    is_mysql = 'mysql' in engine.dialect.dialect_description
    dtype = {'profileid':VARCHAR(10)} if is_mysql else None
    
    # mix crappy letters with a few normal ones
    crap_char_seq = """/_- ?§$&"',:;*()%[]{}|<>=!+#""" + "\\" + "sknalji"  

    # add columns with crappy names
    # don't do this for MySQL which has more strict rules for column names 
    if not is_mysql:
        for i in range(5):
            random_crappy_col_name = ''.join([random.choice(crap_char_seq)
                                              for i in range(50)])

            df_test = (pd.DataFrame({random_crappy_col_name: ['test', None]})
                       .rename_axis(['profileid'], axis='index', inplace=False))

            # psycopg2 can't process columns with "%" or "(" or ")"
            df_test = fix_psycopg2_bad_cols(df_test)
            upsert(engine=engine, schema=schema, df=df_test, if_row_exists='update', dtype=dtype, **default_args)

    # add crappy text in a column named 'text'
    create_random_text = lambda: ''.join([random.choice(crap_char_seq)
                                          for i in range(10)])

    df_test = (pd.DataFrame({'text': [create_random_text() for i in range(10)]})
               .rename_axis(['profileid'], axis='index', inplace=False))
    upsert(engine=engine, schema=schema, df=df_test, if_row_exists='update', dtype=dtype, **default_args)



"""
Functions for preparing/compiling and executing upsert statements
in different SQL flavors.
"""
from copy import deepcopy
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.dialects.mysql.dml import insert as mysql_insert

# # Postgres

def postgres_upsert(engine, connection, table, values, if_row_exists):
    """
    Prepares and executes a PostgreSQL INSERT ON CONFLICT...DO NOTHING
    or DO UPDATE statement via sqlalchemy.dialects.postgresql.insert

    Parameters
    ----------
    engine : sqlalchemy.engine.base.Engine
    connection : sqlalchemy.engine.base.Connection
    table : sqlalchemy.sql.schema.Table
    values : list of dict
    if_row_exists : {'update', 'ignore'}
        * If 'update' issues a ON CONFLICT...DO UPDATE statement
        * If 'ignore' issues a ON CONFLICT...DO NOTHING statement
    """
    insert_stmt = pg_insert(table).values(values)
    if if_row_exists == 'ignore':
        upsert = insert_stmt.on_conflict_do_nothing()
    elif if_row_exists == 'update':
        update_cols = [c.name
                       for c in table.c
                       if c not in list(table.primary_key.columns)]
        upsert = insert_stmt.on_conflict_do_update(index_elements=table.primary_key.columns,
                                                   set_={k: getattr(insert_stmt.excluded, k)
                                                         for k in update_cols})        
    # execute upsert
    connection.execute(upsert)

# # MySQL

def mysql_upsert(engine, connection, table, values, if_row_exists):
    """
    Prepares and executes a MySQL INSERT IGNORE or
    INSERT...ON DUPLICATE KEY UPDATE
    via sqlalchemy.dialects.mysql.dml.insert

    Parameters
    ----------
    engine : sqlalchemy.engine.base.Engine
    connection : sqlalchemy.engine.base.Connection
    table : sqlalchemy.sql.schema.Table
    values : list of dict
    if_row_exists : {'update', 'ignore'}
        * If 'update' issues a ON DUPLICATE KEY UPDATE statement
        * If 'ignore' issues a INSERT IGNORE statement
    """
    insert_stmt = mysql_insert(table).values(values)
    if if_row_exists == 'ignore':
        # thanks to: https://stackoverflow.com/a/50870348/10551772
        upsert = insert_stmt.prefix_with('IGNORE')
    elif if_row_exists == 'update':
        # thanks to: https://stackoverflow.com/a/58180407/10551772
        # prepare kwargs for on_duplicated_key_update (with kwargs and getattr
        # even "bad" column names will resolve e.g. columns with spaces)
        update_cols = {}
        for col in insert_stmt.table.columns:
            col_name = col.name
            if col_name not in table.primary_key:
                update_cols.update({col_name:getattr(insert_stmt.inserted, col_name)})
        upsert = insert_stmt.on_duplicate_key_update(**update_cols)
    # execute upsert
    connection.execute(upsert)

# # Sqlite 
# (or other databases where ON CONFLICT...DO UPDATE SET/DO NOTHING is supported)

def sqlite_upsert(engine, connection, table, values, if_row_exists):
    """
    Compiles and executes a SQlite ON CONFLICT...DO NOTHING or DO UPDATE
    statement.

    Parameters
    ----------
    engine : sqlalchemy.engine.base.Engine
    connection : sqlalchemy.engine.base.Connection
    table : sqlalchemy.sql.schema.Table
    values : list of dict
    if_row_exists : {'update', 'ignore'}
        * If 'update' issues a ON CONFLICT...DO UPDATE statement
        * If 'ignore' issues a ON CONFLICT...DO NOTHING statement

    Examples
    --------
    >>> import datetime
    >>> from sqlalchemy import create_engine
    >>> from pangres.examples import _TestsExampleTable
    >>> from pangres.helpers import PandasSpecialEngine
    >>> 
    >>> engine = create_engine('sqlite:///:memory:')
    >>> df = _TestsExampleTable.create_example_df(nb_rows=5)
    >>> df # doctest: +SKIP
    | profileid   | email             | timestamp                 |   size_in_meters | likes_pizza   | favorite_colors              |
    |:------------|:------------------|:--------------------------|-----------------:|:--------------|:-----------------------------|
    | abc0        | foobaz@gmail.com  | 2007-10-11 23:15:06+00:00 |          1.93994 | False         | ['yellow', 'blue']           |
    | abc1        | foobar@yahoo.com  | 2007-11-21 07:18:20+00:00 |          1.98637 | True          | ['blue', 'pink']             |
    | abc2        | foobaz@outlook.fr | 2002-09-30 17:55:09+00:00 |          1.55945 | True          | ['blue']                     |
    | abc3        | abc@yahoo.fr      | 2007-06-13 22:08:36+00:00 |          2.2495  | True          | ['orange', 'blue']           |
    | abc4        | baz@yahoo.com     | 2004-11-22 04:54:09+00:00 |          2.2019  | False         | ['orange', 'yellow', 'blue'] |

    >>> pse = PandasSpecialEngine(engine=engine, df=df, table_name='test_upsert_sqlite')
    >>> 
    >>> insert_values = {'profileid':'abc5', 'email': 'toto@gmail.com',
    ...                  'timestamp': datetime.datetime(2019, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
    ...                  'size_in_meters':1.9,
    ...                  'likes_pizza':True,
    ...                  'favorite_colors':['red', 'pink']}
    >>> 
    >>> sqlite_upsert(engine=engine, connection=engine.connect(), table=pse.table,
    ...               values=list(insert_values.values()), if_row_exists='update')  # doctest: +SKIP
    """
    def escape_col(col):
        # unbound column from its table
        # otherwise the column would compile as "table.col_name"
        # which we could not use in e.g. SQlite
        unbound_col = deepcopy(col)
        unbound_col.table = None
        return str(unbound_col.compile(dialect=engine.dialect))

    # prepare start of insert (INSERT VALUES (...) ON CONFLICT)
    pk = [escape_col(c) for c in table.primary_key]
    insert = SQLCompiler(dialect=engine.dialect,
                         statement=table.insert().values(values))

    # append on conflict clause
    pk = [escape_col(c) for c in table.primary_key]
    ondup = f'ON CONFLICT ({",".join(pk)})'
    if if_row_exists == 'ignore':
        ondup_action = 'DO NOTHING'
        insert.string = ' '.join((insert.string, ondup, ondup_action))
    elif if_row_exists == 'update':
        ondup_action = 'DO UPDATE SET'
        non_pks = [escape_col(c) for c in table.columns
                   if c not in list(table.primary_key)]
        updates = ', '.join(f'{c}=EXCLUDED.{c}' for c in non_pks)
        insert.string = ' '.join((insert.string, ondup, ondup_action, updates))
    connection.execute(insert)
